import threading
from typing import Optional

from paramiko.channel import ChannelStdinFile, ChannelFile, ChannelStderrFile


class RemoteCommand:
    """
    RemoteCommand wraps the IO pipes of a paramiko command. It is with-able and
    uses threads to guarantee easy IO. Note that upon exit of the with-statement,
    the RemoteCommand will be aborted if it has not already exit. Use the RemoteCommand.wait
    function to wait for the command exit. Note that if you do not catch the TimeoutError
    within the with-statement, the __exit__ call will cause an abortion of the command.
    """
    def __init__(self, stdin: ChannelStdinFile, stdout: ChannelFile, stderr: ChannelStderrFile):
        self.stdin = stdin
        self._stdout = stdout
        self._stderr = stderr

        self._stdout_buff_lock = threading.Lock()
        self._stdout_buff = ""
        self._stderr_buff_lock = threading.Lock()
        self._stderr_buff = ""

        self._stdout_consumer_ready = threading.Event()
        self._stderr_consumer_ready = threading.Event()

        self._stdout_consumer = threading.Thread(target=self._consume_stdout)
        self._stderr_consumer = threading.Thread(target=self._consume_stderr)
        self._stdout_consumer.start()
        self._stderr_consumer.start()

        self._stdout_consumer_ready.wait()
        self._stderr_consumer_ready.wait()

        self._exit_code: Optional[int] = None
        self._has_exit = threading.Event()
        self._exit_listener = threading.Thread(target=self._wait)
        self._exit_listener.start()

    def _consume_stdout(self):
        try:
            self._stdout_consumer_ready.set()
            while True:
                line = self._stdout.readline()
                if not line:
                    break
                with self._stdout_buff_lock:
                    self._stdout_buff += line
        except:
            return

    def _consume_stderr(self):
        try:
            self._stderr_consumer_ready.set()
            while True:
                line = self._stderr.readline()
                if not line:
                    break
                with self._stderr_buff_lock:
                    self._stderr_buff += line
        except:
            return

    def _wait(self):
        self._exit_code = self._stdout.channel.recv_exit_status()
        self._has_exit.set()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    @property
    def exit_code(self) -> Optional[int]:
        """
        Returns the exit_code of the command or None if the command is still running.
        In case of a failure or an aborted command, -1 will be returned as paramiko does via
        recv_exit_status.
        :return: Exit code of the command or None if the command is still running.
        """
        if not self._has_exit.is_set():
            return None
        return self._exit_code

    def wait(self, timeout: Optional[float] = None) -> int:
        """
        Waits for the command to exit. Raises a TimeoutError if the timeout is reached
        before the command has exit. The command will not be aborted automatically if this happens.
        :param timeout: Timeout in (fractions of) seconds. Infinity if None.
        :return: The exit code of the process.
        """
        self._has_exit.wait(timeout)
        if self.exit_code is None:
            raise TimeoutError("waiting for command exit timed out")
        return self.exit_code

    @property
    def stdout(self) -> str:
        """
        Returns the asynchronously captured stdout.
        :return: The captured stdout.
        """
        with self._stdout_buff_lock:
            return self._stdout_buff

    @property
    def stderr(self) -> str:
        """
        Returns the asynchronously captured stderr.
        :return:
        """
        with self._stderr_buff_lock:
            return self._stderr_buff

    def close(self):
        """
        Closes the command, aborting it if it has not exit yet.
        """
        if not self.stdin.closed:
            self.stdin.close()
            self.stdin.channel.close()
        if not self._stdout.closed:
            self._stdout.close()
        if not self._stderr.closed:
            self._stderr.close()
        self._stdout_consumer.join()
        self._stderr_consumer.join()
        self._exit_listener.join()



import os
import socket
from typing import Optional

import paramiko as paramiko
from paramiko import SSHException
from paramiko.sftp_client import SFTPClient
from paramiko.ssh_exception import NoValidConnectionsError

from .command import RemoteCommand


class MaxAttemptsExceededError(RuntimeError):
    pass


class SSHConnection:
    """
    SSHConnection is a wrapper for a paramiko.SSHClient with sane defaults
    for zero-user-interaction automation.
    """

    def __init__(self, host: str,
                 username: Optional[str], password: Optional[str] = None, key_filename: Optional[str] = None,
                 timeout: Optional[float] = None, max_attempts_count: Optional[int] = 1, auto_add_host_key: bool = True, **kwargs):
        """
        Initializes and connects using paramiko and sane defaults for zero-user-interaction automation.

        :param host: Hostname or IP to connect to.
        :param username: Username used for login.
        :param password: Optional password used for login (if omitted, key_filename must be given).
        :param key_filename: Optional key filename used for login (if omitted, password must be given).
        :param timeout: Optional timeout in seconds for connection.
        :param max_attempts_count: Max. number of connection attempts, infinite if None.
        :param auto_add_host_key: If True (default), the missing host key policy "AutoAdd" will be used.
        :param kwargs: Additional args passed to paramiko.SSHClient.connect.
        """
        if "allow_agent" not in kwargs:
            # Prevents interaction with the ssh-agent, which is safer.
            kwargs["allow_agent"] = False
        if "look_for_keys" not in kwargs:
            # Prevents looking for keys by default. This is useful in case of automation
            # alongside a fallback password. In this case you typically don't want paramiko
            # to go looking for any old key on your system, which might trigger a password
            # prompt.
            kwargs["look_for_keys"] = False
        if key_filename:
            kwargs["key_filename"] = key_filename
        if password:
            kwargs["password"] = password

        if not key_filename and not password:
            raise ValueError("at least one of key_filename or password must be given")

        self._ssh = paramiko.SSHClient()
        if auto_add_host_key:
            self._ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())

        attempts = 0
        while True:
            try:
                attempts += 1
                self._ssh.connect(hostname=host, username=username, timeout=timeout, **kwargs)
                break
            except (NoValidConnectionsError, socket.error) as e:
                if max_attempts_count and attempts >= max_attempts_count:
                    raise MaxAttemptsExceededError(f'exceeded number of connection attempts: {max_attempts_count}') from e

    @property
    def client(self) -> paramiko.SSHClient:
        return self._ssh

    def exec(self, command: str) -> RemoteCommand:
        """
        Executes a command and returns the RemoteCommand object, which will handle IO.
        :param command: The command line to be executed. Note that escaping of arguments may depend on the shell used and is up to the user.
        :return: The RemoteCommand object to handle IO.
        """
        stdin, stdout, stderr = self._ssh.exec_command(command)
        return RemoteCommand(stdin, stdout, stderr)

    def upload_recursive(self, local_dir: str, remote_dir: str):
        """
        Recursively traverses the given local_dir and uploads the entire tree structure with all files
        to the remote_dir.
        :param local_dir: Local path to be uploaded.
        :param remote_dir: Remote path to be uploaded to.
        """
        local_dir = local_dir.rstrip("/")
        remote_dir = remote_dir.rstrip("/")

        with self._ssh.open_sftp() as sftp:
            sftp: SFTPClient
            for path, subdirs, files in os.walk(local_dir):
                subpath = path[len(local_dir):].lstrip("/")
                remotepath = f"{remote_dir}/{subpath}".rstrip("/")

                for subdir in subdirs:
                    try:
                        sftp.mkdir(f"{remotepath}/{subdir}")
                    except:
                        pass
                for file in files:
                    sftp.put(f"{path}/{file}", f"{remotepath}/{file}")

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def close(self):
        """
        Closes the connection.
        """
        self._ssh.close()
